/*========================================================================*\

Copyright (c) 1996, 1997  Paul Vojta

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
PAUL VOJTA BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

\*========================================================================*/

/*
 *	hpvptyd - pty daemon for Hewlett-Packard networked printers.
 *
 *	Basically it just uses standard asynchronous I/O techniques to tie
 *	together /dev/pty?? and an Internet socket.
 *	It only closes the connection, though, after the /dev/tty?? has closed
 *	its file descriptor, and after <timeout> seconds have passed.
 *	Some of the more severe errors are logged by syslog(3).
 *
 *	Usage: hpvptyd -m master_pty -x peripheral [args]
 *	where the additional args are:
 *		-p port		Internet port number of the peripheral
 *		-l logfile	Send detailed log to logfile
 *		-l -		Send detailed log to stdout, and refrain from
 *				running itself in the background
 *		-q		Quiet:  don't give number of bytes received
 *				or sent each time.
 *
 *	Compile with gcc.
 *
 *	Written by Paul Vojta
 */

#define	__USE_FIXED_PROTOTYPES__
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <stdarg.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <syslog.h>

extern	int	errno;
extern	int	sys_nerr;
extern	char	*sys_errlist[];

typedef	char	Boolean;
#define	False	0
#define	True	1

/*	Command-line stuff */

const char		*ptyname	= NULL;
const char		*hostname	= NULL;
int			timeout		= 30;
const char		*logfile	= NULL;
int			verbosity	= 1;

int			ptyfd;
int			sock;
struct sockaddr_in	addr;
int			logfd;
pid_t			our_pid;
char			logline[128];
char			*logbody;

/*	Signal flags and select() stuff */

sigset_t		oursigset;
sigset_t		oldsig;
sigset_t		sigalrmset;
volatile Boolean	got_sigpoll;
volatile Boolean	got_sigterm;
volatile Boolean	got_sigalrm;

#define	ALRM_NOP	'\0'
#define	ALRM_CONN	(char) 1
#define	ALRM_TIMEOUT	(char) 2
char			why_alarm	= ALRM_NOP;

fd_set		readfds;
fd_set		writefds;
fd_set		exceptfds;

int		numfds;
struct timeval	s_timeout;

/*	I/O buffers and control variables */

#ifndef	BUF1SIZ
#define	BUF1SIZ	8192
#endif

#ifndef	BUF2SIZ
#define	BUF2SIZ	1024
#endif

char	buf1[BUF1SIZ];		/* downstream buffer (to printer) */
char	*buf1pos	= NULL;	/* NULL or next character to send */
char	*buf1end	= buf1;	/* buf1 or last + 1 character to send */

char	buf2[BUF2SIZ];		/* upstream buffer (from printer) */
char	*buf2pos	= NULL;	/* NULL or next character to send */
char	*buf2end	= buf2;	/* buf2 or last + 1 character to send */

/*	Other */

Boolean	connected	= False;/* Are we connected to the socket? */
Boolean	pty_err		= False;/* Was the pty recently closed? */

#define	LOG_LOCAL	0x8000	/* bigger than any LOG_* defined in syslog.h */
#define	LOG_VERBOSE	(LOG_LOCAL | 1)
int	connect_errno	= 0;	/* errno on last connect */

/*
 *	Log file handler
 */

void
logprintf(int priority, const char *format, ...)
{
	va_list		args;
	time_t		t;
	char		*p;

	if (logfile == NULL && (priority & LOG_LOCAL))
	    return;

	/* Form the message body */

	va_start(args, format);
	vsprintf(logbody, format, args);
	va_end(args);
	p = logbody + strlen(logbody);

	/* Put it on the log file */

	if (logfile != NULL) {
	    if ((priority & LOG_LOCAL) && (priority & 7) > verbosity)
		return;

	    t = time(NULL);
	    memcpy(logline, ctime(&t) + 11, 9);
	    *p = '\n';
	    write(logfd, logline, p - logline + 1);
	}

	/* Put it on syslog */

	if (priority & LOG_LOCAL)
	    return;

	syslog(priority, logbody);
}

void
logperror(int priority, const char *why)
{
	connect_errno = 0;
	if (errno > 0 && errno < sys_nerr)
	    logprintf(priority, "%s:  %s", why, sys_errlist[errno]);
	else
	    logprintf(priority, "%s:  error number %d", why, errno);
}

/*
 *	Signal handlers.
 */

void
handle_sigpoll(int signo)
{
	got_sigpoll = True;
}

void
handle_sigterm(int signo)
{
	got_sigterm = True;
}

void
handle_sigalrm(int signo)
{
	got_sigalrm = True;
}

void
cancel_alarm(void)
{
	alarm(0);

	/* Allow any pending SIGALRM to occur */
	sigprocmask(SIG_UNBLOCK, &sigalrmset, NULL);
	sigprocmask(SIG_BLOCK, &sigalrmset, NULL);

	got_sigalrm = False;
	why_alarm = ALRM_NOP;
}


/*
 *	Connect to socket.
 */

void
try_to_connect(void)
{
	sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (sock == -1) {
	    logperror(LOG_ERR, "socket");
	    exit(1);
	}

	if (connect(sock, (struct sockaddr *) &addr, sizeof(addr)) == -1) {
	    logperror(errno == connect_errno ? LOG_LOCAL : LOG_NOTICE,
		"connect");
	    connect_errno = errno;
	    why_alarm = ALRM_CONN;
	    alarm(15);		/* try again later */
	    close(sock);
	    return;
	}

	/* Set socket for non-blocking I/O */

	if (fcntl(sock, F_SETOWN, getpid()) == -1)
	    logperror(LOG_ERR, "fcntl F_SETOWN (socket)");

	if (fcntl(sock, F_SETFL,
		fcntl(sock, F_GETFL, 0) | FASYNC | O_NONBLOCK) == -1)
	    logperror(LOG_ERR, "fcntl F_SETFL (socket)");

	numfds = ptyfd;
	if (sock > numfds) numfds = sock;
	++numfds;

	logprintf(LOG_LOCAL, "Connected");
	connected = True;
	got_sigpoll = True;	/* go through the loop again */

	if (pty_err) {
	    why_alarm = ALRM_TIMEOUT;
	    alarm(timeout);
	}
}

/*
 *	Disconnect socket.
 */

void
disconnect(int priority, const char *why)
{
	cancel_alarm();

	if (!connected)
	    return;

	close(sock);
	connected = False;
	buf1pos = NULL;
	buf1end = buf1;
	logprintf(priority, why);
}


/*
 *	Main program.
 */

int
main(int argc, char **argv)
{
	extern	const char	*optarg;
	int			opt;
	Boolean			errflg	= False;
	struct sigaction	a;
	struct hostent		*hostrecp;
	int			n;

	/*
	 * Process arguments
	 */

	addr.sin_port = 9100;

	while ((opt = getopt(argc, argv, "m:x:p:t:l:q")) != -1) {
	    switch (opt) {
		case 'm':  ptyname = optarg; break;
		case 'x':  hostname = optarg; break;
		case 'p':  addr.sin_port = atoi(optarg); break;
		case 't':  timeout = atoi(optarg); break;
		case 'l':  logfile = optarg; break;
		case 'q':  verbosity = 0; break;
		case '?':  errflg = True;
	    }
	}

	if (errflg || ptyname == NULL || hostname == NULL) {
	    fputs("Usage:  hpvptyd -m master_pty -x peripheral [-p port] "
		"[-t timeout]\n\t\t[-l logfile] [-q]\n", stderr);
	    exit(1);
	}

	/*
	 * Resolve hostname
	 */

	errno = 0;
	hostrecp = gethostbyname(hostname);
	if (hostrecp == NULL) {
	    if (errno == EACCES || errno == EINVAL)
		fprintf(stderr, "%s:  Host unknown", hostname);
	    else
		perror(hostname);
	    exit(1);
	}

	addr.sin_family = htons(AF_INET);
	if (hostrecp->h_length != 4 || hostrecp->h_addr_list == NULL) {
	    fprintf(stderr, "%s:  address missing or of wrong length",
		hostname);
	    exit(1);
	}
	memcpy(&addr.sin_addr.s_addr, *(hostrecp->h_addr_list), 4);

	/*
	 * Set up log file and (maybe) fork a subprocess.
	 */

	if (logfile == NULL) {
	    if ((our_pid = fork()) > 0)
		return 0;
	}
	else if (strcmp(logfile, "-") == 0)
	    logfd = 1;	/* stdout */
	else {
	    logfd = open(logfile, O_WRONLY | O_APPEND | O_CREAT, 0644);
	    if (logfd == -1) {
		perror(logfile);
		exit(1);
	    }
	    if ((our_pid = fork()) > 0)
		return 0;
	}

	if (our_pid == -1) {
	    perror("fork");
	    exit(1);
	}

	our_pid = getpid();
	sprintf(logline + 9, "hpvptyd[%d] ", our_pid);
	logbody = logline + 9 + strlen(logline + 9);
	openlog("hpvptyd", LOG_PID, LOG_LPR);

	logprintf(LOG_LOCAL, "Started (-m %s -x %s -p %d -t %d)", ptyname,
	    hostname, addr.sin_port, timeout);

	/*
	 * Set up signal handlers
	 */

	sigemptyset(&oursigset);
	sigaddset(&oursigset, SIGPOLL);
	sigaddset(&oursigset, SIGALRM);
	sigaddset(&oursigset, SIGTERM);
	sigaddset(&oursigset, SIGQUIT);
	sigaddset(&oursigset, SIGINT);
	a.sa_mask = oursigset;

	a.sa_handler = handle_sigpoll;
	a.sa_flags = 0;
	sigaction(SIGPOLL, &a, NULL);

	a.sa_handler = SIG_IGN;
	sigaction(SIGPIPE, &a, NULL);

	a.sa_handler = handle_sigalrm;
	sigaction(SIGALRM, &a, NULL);

	a.sa_handler = handle_sigterm;
	a.sa_flags = SA_RESETHAND;
	sigaction(SIGTERM, &a, NULL);
	sigaction(SIGQUIT, &a, NULL);
	sigaction(SIGINT, &a, NULL);

	sigprocmask(SIG_BLOCK, &oursigset, &oldsig);

	sigemptyset(&sigalrmset);
	sigaddset(&sigalrmset, SIGALRM);

	/*
	 * Open the pseudo-terminal
	 */

	ptyfd = open(ptyname, O_RDWR | O_NONBLOCK);
	if (ptyfd == -1) {
	    logperror(LOG_ERR, ptyname);
	    exit(1);
	}

	if (fcntl(ptyfd, F_SETOWN, getpid()) == -1) {
	    logperror(LOG_ERR, "fcntl F_SETOWN");
	    exit(1);
	}

	if (fcntl(ptyfd, F_SETFL,
		fcntl(ptyfd, F_GETFL, 0) | FASYNC) == -1) {
	    logperror(LOG_ERR, "fcntl F_SETFL");
	    exit(1);
	}

	numfds = ptyfd + 1;

	/*
	 * Main loop
	 */

	for (;;) {
	    /*
	     *	I've had problems with SIGPOLL being generated correctly;
	     *	in particular, if I write to the socket, but not all the
	     *	bytes get sent, then I don't necessarily get a SIGPOLL when
	     *	it is possible to send more.  So, after any successful I/O,
	     *	we go around again.
	     */
	    if (got_sigpoll) {
		/* We're going around again, so allow signals to happen. */
		sigprocmask(SIG_SETMASK, &oldsig, NULL);
		sigprocmask(SIG_BLOCK, &oursigset, NULL);
	    }
	    else
		sigsuspend(&oldsig);	/* wait for something to happen */

	    if (got_sigterm) {
		logprintf(LOG_NOTICE, "Terminated");
		break;
	    }

	    if (got_sigalrm) {
		got_sigalrm = False;
		switch (why_alarm) {
		    case ALRM_CONN:
			try_to_connect();
			break;
		    case ALRM_TIMEOUT:
			disconnect(LOG_LOCAL, "Socket timed out");
			break;
		}
	    }

	    if (got_sigpoll) {
		got_sigpoll = False;

		FD_ZERO(&readfds);
		FD_ZERO(&writefds);
		FD_ZERO(&exceptfds);

		if (buf1pos == NULL || buf1end < buf1 + BUF1SIZ)
		    FD_SET(ptyfd, &readfds);
		FD_SET(ptyfd, &writefds);
		if (!pty_err) FD_SET(ptyfd, &exceptfds);

		if (connected) {
		    if (buf2pos == NULL || buf2end < buf2 + BUF2SIZ)
			FD_SET(sock, &readfds);
		    FD_SET(sock, &writefds);
		    FD_SET(sock, &exceptfds);
		}

		memset((char *) &s_timeout, '\0', sizeof(s_timeout));

		n = select(numfds, &readfds, &writefds, &exceptfds, &s_timeout);
		if (n < 0) {
		    logperror(LOG_ERR, "select");
		    continue;
		}

		if (FD_ISSET(ptyfd, &exceptfds)) {
		    logprintf(LOG_LOCAL, "Got exception mention on pty");
		    pty_err = True;
		    if (connected) {
			why_alarm = ALRM_TIMEOUT;
			alarm(timeout);
		    }
		}
		if (FD_ISSET(sock, &exceptfds))
		    logprintf(LOG_NOTICE, "Got exception mention on socket");

		if (FD_ISSET(ptyfd, &readfds)) {
		    n = read(ptyfd, buf1end, buf1 + BUF1SIZ - buf1end);
		    if (n < 0) {
			if (pty_err && errno == EIO) {
			    /* clear upstream data */
			    buf2pos = NULL;
			    buf2end = buf2;
			}
			else
			    logperror(LOG_ERR, "read (pty)");
		    }
		    else {
			logprintf(LOG_VERBOSE, "Received %d bytes from pty", n);
			got_sigpoll = True;	/* go around again */
			if (n > 0) {
			    if (pty_err) {
				pty_err = False;
				if (why_alarm == ALRM_TIMEOUT)
				    cancel_alarm();
			    }
			    if (buf1pos == NULL) {
				buf1pos = buf1;
				if (!connected)
				    try_to_connect();
			    }
			    buf1end += n;
			}
		    }
		}

		if (connected && buf1pos != NULL && FD_ISSET(sock, &writefds)) {
		    n = write(sock, buf1pos, buf1end - buf1pos);
		    if (n < 0) {
			if (errno == EPIPE)
			    disconnect(LOG_INFO, "Socket closed by printer");
			else
			    logperror(LOG_ERR, "write (socket)");
		    }
		    else {
			logprintf(LOG_VERBOSE, "Sent %d bytes to socket", n);
			got_sigpoll = True;	/* go around again */
			buf1pos += n;
			if (buf1pos == buf1end) {
			    buf1pos = NULL;
			    buf1end = buf1;
			}
			if (why_alarm == ALRM_TIMEOUT)
			    alarm(timeout);	/* reset timeout */
		    }
		}

		if (connected && FD_ISSET(sock, &readfds)) {
		    n = read(sock, buf2end, buf2 + BUF2SIZ - buf2end);
		    if (n < 0) {
			if (errno == ECONNRESET)
			    disconnect(LOG_INFO, "Socket closed by printer");
			else
			    logperror(LOG_ERR, "read (socket)");
		    }
		    else {
			logprintf(LOG_VERBOSE, "Received %d bytes from socket",
			    n);
			got_sigpoll = True;	/* go around again */
			if (n > 0) {
			    if (buf2pos == NULL)
				buf2pos = buf2;
			    buf2end += n;
			    if (why_alarm == ALRM_TIMEOUT)
				alarm(timeout);		/* reset timeout */
			}
		    }
		}

		if (buf2pos != NULL && FD_ISSET(ptyfd, &writefds)) {
		    n = write(ptyfd, buf2pos, buf2end - buf2pos);
		    if (n < 0)
			logperror(LOG_ERR, "write (pty)");
		    else {
			logprintf(LOG_VERBOSE, "Sent %d bytes to pty", n);
			got_sigpoll = True;	/* go around again */
			buf2pos += n;
			if (buf2pos == buf2end) {
			    buf2pos = NULL;
			    buf2end = buf2;
			}
		    }
		}
	    }
	}

	closelog();
	return 0;
}
